<?php

namespace apexl\Io\modules\user\entities;

use apexl\Io\includes\Entity;

class permissionsEntity extends Entity {

    public function __construct()
    {
        parent::__construct('permissions');
    }

    public function store(){
        parent::__call('store', [$this->data]);
        return $this;
    }
}